#!/usr/bin/env python
#-*- coding:utf-8 -*-

# Copyright (C) 2010 Antony Templier
# 
# This software is under WTFPL License:
# ------------------------------------------------------------------------------
#             DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#                     Version 2, December 2004
# 
#  Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
# 
#  Everyone is permitted to copy and distribute verbatim or modified
#  copies of this license document, and changing it is allowed as long
#  as the name is changed.
# 
#             DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#    TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
# 
#   0. You just DO WHAT THE FUCK YOU WANT TO.
# ------------------------------------------------------------------------------

# This is just a rewrite in python of the algorithm used to make to the history 
# part of the pages found in the phpdiff.php script of a standard PmWiki 
# distributuon : <http://www.pmwiki.org>.

# Found in the original script distributed under the General Public License :
# Copyright 2003,2004 Nils Knappmeier (nk@knappi.org)
# Copyright 2004-2010 Patrick R. Michaud (pmichaud@pobox.com)

# itself based on Daniel Unterberger's diff 
# <http://www.holomind.de/phpnet/diff.php>

"""
Python version of default diff algorithm of PmWiki
"""

__version__ = '1.0.0'
__author__ = 'Antony Templier <antony.templier@gmail.com>'
__date__ = "30 june 2011"
__license__ = "WTFPL"
__credits__ = 'Original algorithm by Nils Knappmeier & Partick R. Michaud.'


def diff(old, new):

    t1 = old.split(u'\n')
    x = t1.pop()
    if x > '':
        t1.append(u'{0}\n\\ No newline at end of file'.format(x))

    t2 = new.split(u'\n')
    x = t2.pop()
    if x > '':
        t2.append(u'{0}\n\\ No newline at end of file'.format(x))

    t1_start, t1_end = 0, len(t1)
    t2_start, t2_end = 0, len(t2)

    # stop with a common ending
    while t1_start < t1_end and t2_start < t2_end \
        and t1[t1_end-1] == t2[t2_end-1]:
        t1_end -= 1
        t2_end -= 1

    # skip over any common beginning
    while t1_start < t1_end and t2_start < t2_end \
        and t1[t1_start] == t2[t2_start]:
        t1_start += 1
        t2_start += 1

    # build a reverse-index array using the line as key and line number as value
    # don't store blank lines, so they won't be targets of the shortest distance
    # search
    r1 = {}
    for i in range(t1_start, t1_end):
        if t1[i] > '':
            if not t1[i] in r1:
                r1[t1[i]] = [i]
            else:
                r1[t1[i]].append(i)

    r2 = {}
    for i in range(t2_start, t2_end):
        if t2[i] > '':
            if not t2[i] in r2:
                r2[t2[i]] = [i]
            else:
                r2[t2[i]].append(i)

    a1 = t1_start # start at beginning of each list
    a2 = t2_start
    actions = []

    # walk this loop until we reach the end of one of the lists
    while a1 < t1_end and a2 < t2_end:

        # if we have a common element, save it and go to the next
        if t1[a1] == t2[a2]:
            actions.append(4)
            a1 += 1
            a2 += 1
            continue

        # otherwise, find the shortest move (Manhattan-distance) from the
        # current location
        best1 = t1_end
        best2 = t2_end
        s1 = a1
        s2 = a2

        while (s1 + s2 - a1 - a2) < (best1 + best2 - a1 - a2):

            d = -1

            try:
                for n in r1[t2[s2]]:
                    if n >= s1:
                        d = n
                        break
            except:
                pass

            if d >= s1 and (d + s2 - a1 - a2) < (best1 + best2 - a1 - a2):
                best1 = d
                best2 = s2
            d = -1

            try:
                for n in r2[t1[s1]]:
                    if n >= s2:
                        d = n
                        break
            except:
                pass

            if d >= s2 and (s1 + d - a1 - a2) < (best1 + best2 - a1 - a2):
                best1 = s1
                best2 = d
            s1 += 1
            s2 += 1

        while a1 < best1: # deleted elements
            actions.append(1)
            a1 += 1
        while a2 < best2: # added elements
            actions.append(2)
            a2 +=1

    # we've reached the end of one list, now walk to the end of the other
    while a1 < t1_end: # deleted elements
        actions.append(1)
        a1 += 1
    while a2 < t2_end: # deleted elements
        actions.append(2)
        a2 += 1

    # and this marks our ending point
    actions.append(8)

    # now, let's follow the path we just took and report the added/deleted
    # elements into $out.
    op = 0
    x0 = x1 = t1_start
    y0 = y1 = t2_start
    out = []

    for act in actions:
        if act == 1:
            op |= act
            x1 += 1
            continue
        if act == 2:
            op |= act
            y1 += 1
            continue
        if op > 0:
            xstr = x1
            if not x1 == (x0 + 1):
                xstr = u'{0},{1}'.format( (x0 + 1), x1)
            ystr = y1
            if not y1 == (y0 + 1):
                ystr = u'{0},{1}'.format( (y0 + 1), y1)
            if op == 1:
                out.append(u'{0}d{1}'.format(xstr, y1))
            elif op == 3:
                out.append(u'{0}c{1}'.format(xstr, ystr))
            while x0 < x1:
                out.append(u'< {0}'.format(t1[x0])) # deleted elems
                x0 += 1
            if op == 2:
                out.append(u'{0}a{1}'.format(x1, ystr))
            elif op == 3:
                out.append(u'---')
            while y0 < y1:
                out.append(u'> {0}'.format(t2[y0])) # added elems
                y0 += 1
        x1 += 1
        x0 = x1
        y1 += 1
        y0 = y1
        op = 0

    out.append('')
    resultat = '\n'.join(out)

    return resultat
